from random import choices, seed
import numpy as np

import torch
from functools import reduce
from operator import mul

from torch import nn
from torch.optim import Adam

from datasets.trajectory_dataset import TrajData
from lib.evaluation.classification import ROCEvaluation
from lib.modules.blocks import ConvModule, ResidualModule, DeConvModule
from lib.modules.utils import LightningBaseModule, Flatten

import matplotlib.pyplot as plt


class CNNRouteGeneratorModel(LightningBaseModule):

    name = 'CNNRouteGenerator'

    def configure_optimizers(self):
        return Adam(self.parameters(), lr=self.hparams.train_param.lr)

    def training_step(self, batch_xy, batch_nb, *args, **kwargs):
        batch_x, alternative = batch_xy
        generated_alternative, z, mu, logvar = self(batch_x)
        element_wise_loss = self.criterion(generated_alternative, alternative)
        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)

        kld_loss = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
        # Dimensional Resizing TODO: Does This make sense? Sanity Check it!
        # kld_loss /= reduce(mul, self.in_shape)
        # kld_loss *= self.hparams.data_param.dataset_length / self.hparams.train_param.batch_size * 100

        loss = kld_loss + element_wise_loss
        return dict(loss=loss, log=dict(element_wise_loss=element_wise_loss, loss=loss, kld_loss=kld_loss))

    def _test_val_step(self, batch_xy, batch_nb, *args):
        batch_x, _ = batch_xy
        map_array = batch_x[:, 0].unsqueeze(1)
        trajectory = batch_x[:, 1].unsqueeze(1)
        labels = batch_x[:, 2].unsqueeze(1).max(dim=-1).values.max(-1).values

        _, mu, _ = self.encode(batch_x)
        generated_alternative = self.generate(mu)
        return dict(maps=map_array, trajectories=trajectory, batch_nb=batch_nb,  labels=labels,
                    generated_alternative=generated_alternative, pred_label=-1)

    def _test_val_epoch_end(self, outputs, test=False):
        val_restul_dict = self.generate_random()

        from lib.visualization.generator_eval import GeneratorVisualizer
        g = GeneratorVisualizer(**val_restul_dict)
        fig = g.draw()
        self.logger.log_image(f'{self.name}_Output', fig, step=self.global_step)
        plt.clf()

        return dict(epoch=self.current_epoch)

    def on_epoch_start(self):
        self.dataset.seed(self.logger.version)
        # torch.random.manual_seed(self.logger.version)
        # np.random.seed(self.logger.version)

    def validation_step(self, *args):
        return self._test_val_step(*args)

    def validation_epoch_end(self, outputs: list):
        return self._test_val_epoch_end(outputs)

    def test_step(self, *args):
        return self._test_val_step(*args)

    def test_epoch_end(self, outputs):
        return self._test_val_epoch_end(outputs, test=True)

    def __init__(self, *params, issubclassed=False):
        super(CNNRouteGeneratorModel, self).__init__(*params)

        if not issubclassed:
            # Dataset
            self.dataset = TrajData(self.hparams.data_param.map_root, mode='generator_all_in_map',
                                    preprocessed=self.hparams.data_param.use_preprocessed,
                                    length=self.hparams.data_param.dataset_length, normalized=True)
            self.criterion = nn.MSELoss()

        # Additional Attributes                               #
        #######################################################
        self.in_shape = self.dataset.map_shapes_max
        self.use_res_net = self.hparams.model_param.use_res_net
        self.lat_dim = self.hparams.model_param.lat_dim
        self.feature_dim = self.lat_dim * 10
        ########################################################

        # NN Nodes
        ###################################################
        #
        # Utils
        self.activation = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

        #
        # Map Encoder
        self.enc_conv_0 = ConvModule(self.in_shape, conv_kernel=3, conv_stride=1, conv_padding=1,
                                     conv_filters=self.hparams.model_param.filters[0],
                                     use_norm=self.hparams.model_param.use_norm,
                                     use_bias=self.hparams.model_param.use_bias)

        self.enc_res_1 = ResidualModule(self.enc_conv_0.shape, ConvModule, 2, conv_kernel=5, conv_stride=1,
                                        conv_padding=2, conv_filters=self.hparams.model_param.filters[0],
                                        use_norm=self.hparams.model_param.use_norm,
                                        use_bias=self.hparams.model_param.use_bias)
        self.enc_conv_1a = ConvModule(self.enc_res_1.shape, conv_kernel=3, conv_stride=1, conv_padding=0,
                                      conv_filters=self.hparams.model_param.filters[1],
                                      use_norm=self.hparams.model_param.use_norm,
                                      use_bias=self.hparams.model_param.use_bias)
        self.enc_conv_1b = ConvModule(self.enc_conv_1a.shape, conv_kernel=3, conv_stride=2, conv_padding=0,
                                      conv_filters=self.hparams.model_param.filters[1],
                                      use_norm=self.hparams.model_param.use_norm,
                                      use_bias=self.hparams.model_param.use_bias)

        self.enc_res_2 = ResidualModule(self.enc_conv_1b.shape, ConvModule, 2, conv_kernel=5, conv_stride=1,
                                        conv_padding=2, conv_filters=self.hparams.model_param.filters[1],
                                        use_norm=self.hparams.model_param.use_norm,
                                        use_bias=self.hparams.model_param.use_bias)
        self.enc_conv_2a = ConvModule(self.enc_res_2.shape, conv_kernel=5, conv_stride=1, conv_padding=0,
                                      conv_filters=self.hparams.model_param.filters[2],
                                      use_norm=self.hparams.model_param.use_norm,
                                      use_bias=self.hparams.model_param.use_bias)
        self.enc_conv_2b = ConvModule(self.enc_conv_2a.shape, conv_kernel=5, conv_stride=1, conv_padding=0,
                                      conv_filters=self.hparams.model_param.filters[2],
                                      use_norm=self.hparams.model_param.use_norm,
                                      use_bias=self.hparams.model_param.use_bias)

        self.enc_res_3 = ResidualModule(self.enc_conv_2b.shape, ConvModule, 2, conv_kernel=7, conv_stride=1,
                                        conv_padding=3, conv_filters=self.hparams.model_param.filters[2],
                                        use_norm=self.hparams.model_param.use_norm,
                                        use_bias=self.hparams.model_param.use_bias)
        self.enc_conv_3a = ConvModule(self.enc_res_3.shape, conv_kernel=7, conv_stride=1, conv_padding=0,
                                      conv_filters=self.hparams.model_param.filters[2],
                                      use_norm=self.hparams.model_param.use_norm,
                                      use_bias=self.hparams.model_param.use_bias)
        self.enc_conv_3b = ConvModule(self.enc_conv_3a.shape, conv_kernel=7, conv_stride=1, conv_padding=0,
                                      conv_filters=self.hparams.model_param.filters[2],
                                      use_norm=self.hparams.model_param.use_norm,
                                      use_bias=self.hparams.model_param.use_bias)

        self.enc_flat = Flatten(self.enc_conv_3b.shape)
        self.enc_lin_1 = nn.Linear(self.enc_flat.shape, self.feature_dim)

        #
        # Mixed Encoder
        self.enc_lin_2 = nn.Linear(self.feature_dim, self.feature_dim)
        self.enc_norm = nn.BatchNorm1d(self.feature_dim) if self.hparams.model_param.use_norm else lambda x: x

        #
        # Variational Bottleneck
        self.mu = nn.Linear(self.feature_dim, self.lat_dim)
        self.logvar = nn.Linear(self.feature_dim, self.lat_dim)

        #
        # Alternative Generator
        self.gen_lin_1 = nn.Linear(self.hparams.model_param.lat_dim, self.feature_dim)

        self.gen_lin_2 = nn.Linear(self.feature_dim, self.enc_flat.shape)

        self.reshape_to_last_conv = Flatten(self.enc_flat.shape, self.enc_conv_3b.shape)

        self.gen_deconv_1a = DeConvModule(self.enc_conv_3b.shape, self.hparams.model_param.filters[2],
                                          conv_padding=0, conv_kernel=11, conv_stride=1,
                                          use_norm=self.hparams.model_param.use_norm)
        self.gen_deconv_1b = DeConvModule(self.gen_deconv_1a.shape, self.hparams.model_param.filters[2],
                                          conv_padding=0, conv_kernel=9, conv_stride=2,
                                          use_norm=self.hparams.model_param.use_norm)

        self.gen_deconv_2a = DeConvModule(self.gen_deconv_1b.shape, self.hparams.model_param.filters[1],
                                          conv_padding=0, conv_kernel=7, conv_stride=1,
                                          use_norm=self.hparams.model_param.use_norm)
        self.gen_deconv_2b = DeConvModule(self.gen_deconv_2a.shape, self.hparams.model_param.filters[1],
                                          conv_padding=0, conv_kernel=7, conv_stride=1,
                                          use_norm=self.hparams.model_param.use_norm)

        self.gen_deconv_3a = DeConvModule(self.gen_deconv_2b.shape, self.hparams.model_param.filters[0],
                                          conv_padding=1, conv_kernel=5, conv_stride=1,
                                          use_norm=self.hparams.model_param.use_norm)
        self.gen_deconv_3b = DeConvModule(self.gen_deconv_3a.shape, self.hparams.model_param.filters[0],
                                          conv_padding=1, conv_kernel=4, conv_stride=1,
                                          use_norm=self.hparams.model_param.use_norm)

        self.gen_deconv_out = DeConvModule(self.gen_deconv_3b.shape, 1, activation=None,
                                           conv_padding=0, conv_kernel=3, conv_stride=1,
                                           use_norm=self.hparams.model_param.use_norm)

    def forward(self, batch_x):
        #
        # Encode
        z, mu, logvar = self.encode(batch_x)

        #
        # Generate
        alt_tensor = self.generate(z)
        return alt_tensor, z, mu, logvar

    @staticmethod
    def reparameterize(mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def encode(self, batch_x):
        combined_tensor = self.enc_conv_0(batch_x)
        combined_tensor = self.enc_res_1(combined_tensor) if self.use_res_net else combined_tensor
        combined_tensor = self.enc_conv_1a(combined_tensor)
        combined_tensor = self.enc_conv_1b(combined_tensor)
        combined_tensor = self.enc_res_2(combined_tensor) if self.use_res_net else combined_tensor
        combined_tensor = self.enc_conv_2a(combined_tensor)
        combined_tensor = self.enc_conv_2b(combined_tensor)
        combined_tensor = self.enc_res_3(combined_tensor) if self.use_res_net else combined_tensor
        combined_tensor = self.enc_conv_3a(combined_tensor)
        combined_tensor = self.enc_conv_3b(combined_tensor)

        combined_tensor = self.enc_flat(combined_tensor)
        combined_tensor = self.enc_lin_1(combined_tensor)
        combined_tensor = self.enc_lin_2(combined_tensor)

        combined_tensor = self.enc_norm(combined_tensor)
        combined_tensor = self.activation(combined_tensor)
        combined_tensor = self.enc_lin_2(combined_tensor)
        combined_tensor = self.enc_norm(combined_tensor)
        combined_tensor = self.activation(combined_tensor)

        #
        # Parameter and Sampling
        mu = self.mu(combined_tensor)
        logvar = self.logvar(combined_tensor)
        z = self.reparameterize(mu, logvar)
        return z, mu, logvar

    def generate(self, z):
        alt_tensor = self.gen_lin_1(z)
        alt_tensor = self.activation(alt_tensor)
        alt_tensor = self.gen_lin_2(alt_tensor)
        alt_tensor = self.activation(alt_tensor)
        alt_tensor = self.reshape_to_last_conv(alt_tensor)
        alt_tensor = self.gen_deconv_1a(alt_tensor)
        alt_tensor = self.gen_deconv_1b(alt_tensor)
        alt_tensor = self.gen_deconv_2a(alt_tensor)
        alt_tensor = self.gen_deconv_2b(alt_tensor)
        alt_tensor = self.gen_deconv_3a(alt_tensor)
        alt_tensor = self.gen_deconv_3b(alt_tensor)
        alt_tensor = self.gen_deconv_out(alt_tensor)
        # alt_tensor = self.activation(alt_tensor)
        alt_tensor = self.sigmoid(alt_tensor)
        return alt_tensor

    def generate_random(self, n=12):

        samples, alternatives = zip(*[self.dataset.test_dataset[choice]
                                      for choice in choices(range(self.dataset.length), k=n)])
        samples = self._move_to_model_device(torch.stack([torch.as_tensor(x) for x in samples]))
        alternatives = self._move_to_model_device(torch.stack([torch.as_tensor(x) for x in alternatives]))

        return self._test_val_step((samples, alternatives), -9999)


class CNNRouteGeneratorDiscriminated(CNNRouteGeneratorModel):

    name = 'CNNRouteGeneratorDiscriminated'

    def training_step(self, batch_xy, batch_nb, *args, **kwargs):
        batch_x, label = batch_xy

        generated_alternative, z, mu, logvar = self(batch_x)
        map_array, trajectory = batch_x

        map_stack = torch.cat((map_array, trajectory, generated_alternative), dim=1)
        pred_label = self.discriminator(map_stack)
        discriminated_bce_loss = self.criterion(pred_label, label.float().unsqueeze(-1))

        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        kld_loss = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
        # Dimensional Resizing
        kld_loss /= reduce(mul, self.in_shape)

        loss = (kld_loss + discriminated_bce_loss) / 2
        return dict(loss=loss, log=dict(loss=loss,
                                        discriminated_bce_loss=discriminated_bce_loss,
                                        kld_loss=kld_loss)
                    )

    def _test_val_step(self, batch_xy, batch_nb, *args):
        batch_x, label = batch_xy

        generated_alternative, z, mu, logvar = self(batch_x)
        map_array, trajectory = batch_x

        map_stack = torch.cat((map_array, trajectory, generated_alternative), dim=1)
        pred_label = self.discriminator(map_stack)

        discriminated_bce_loss = self.criterion(pred_label, label.float().unsqueeze(-1))
        return dict(discriminated_bce_loss=discriminated_bce_loss, batch_nb=batch_nb,
                    pred_label=pred_label, label=label, generated_alternative=generated_alternative)

    def validation_step(self, *args):
        return self._test_val_step(*args)

    def validation_epoch_end(self, outputs: list):
        return self._test_val_epoch_end(outputs)

    def _test_val_epoch_end(self, outputs, test=False):
        evaluation = ROCEvaluation(plot_roc=True)
        pred_label = torch.cat([x['pred_label'] for x in outputs])
        labels = torch.cat([x['label'] for x in outputs]).unsqueeze(1)
        mean_losses = torch.stack([x['discriminated_bce_loss'] for x in outputs]).mean()

        # Sci-py call ROC eval call is eval(true_label, prediction)
        roc_auc, tpr, fpr = evaluation(labels.cpu().numpy(), pred_label.cpu().numpy(), )
        if test:
            # self.logger.log_metrics(score_dict)
            self.logger.log_image(f'{self.name}_ROC-Curve', plt.gcf(), step=self.global_step)
        plt.clf()

        maps, trajectories, labels, val_restul_dict = self.generate_random()

        from lib.visualization.generator_eval import GeneratorVisualizer
        g = GeneratorVisualizer(maps, trajectories, labels, val_restul_dict)
        fig = g.draw()
        self.logger.log_image(f'{self.name}_Output', fig, step=self.global_step)
        plt.clf()

        return dict(mean_losses=mean_losses, roc_auc=roc_auc, epoch=self.current_epoch)

    def test_step(self, *args):
        return self._test_val_step(*args)

    def test_epoch_end(self, outputs):
        return self._test_val_epoch_end(outputs, test=True)

    @property
    def discriminator(self):
        if self._disc is None:
            raise RuntimeError('Set the Discriminator first; "set_discriminator(disc_model)')
        return self._disc

    def set_discriminator(self, disc_model):
        if self._disc is not None:
            raise RuntimeError('Discriminator has already been set... What are trying to do?')
        self._disc = disc_model

    def __init__(self, *params):
        raise NotImplementedError
        super(CNNRouteGeneratorDiscriminated, self).__init__(*params, issubclassed=True)

        self._disc = None

        self.criterion = nn.BCELoss()

        self.dataset = TrajData(self.hparams.data_param.map_root, mode='just_route', preprocessed=True,
                                length=self.hparams.data_param.dataset_length, normalized=True)
